@@ -4,12 +4,15 @@ from dataclasses import field
 
 import jax
 import jax.numpy as jnp
+import jax.random as jr
+import jax.tree_util as jtu
 import numpy as np
+import numpy.testing as npt
 import pytest
 
 import pytreeclass as pytc
 from pytreeclass.src.tree_indexer import _at_apply, _at_get, _at_reduce, _at_set
-from pytreeclass.src.tree_util import is_treeclass_equal, static_value
+from pytreeclass.src.tree_util import is_treeclass_equal
 
 
 @pytc.treeclass
@@ -18,7 +21,7 @@ class Test:
     b: float
     c: float
     d: jnp.ndarray
-    name: str
+    name: str = pytc.static_field()
 
 
 def test_getter_by_val():
@@ -71,26 +74,24 @@ def test_getter_by_val():
 
 def test_getter_by_param():
 
-    A = Test(10, 20, 30, jnp.array([1, 2, 3, 4, 5]), static_value("A"))
+    A = Test(10, 20, 30, jnp.array([1, 2, 3, 4, 5]), ("A"))
 
     B = A.at[A == "a"].get(array_as_leaves=False)
-    assert is_treeclass_equal(B, Test(10, None, None, None, static_value("A")))
+    assert is_treeclass_equal(B, Test(10, None, None, None, ("A")))
 
     B = A.at[(A == "a") | (A == "b")].get(array_as_leaves=False)
-    assert is_treeclass_equal(B, Test(10, 20, None, None, static_value("A")))
+    assert is_treeclass_equal(B, Test(10, 20, None, None, ("A")))
 
     B = A.at[A == ""].get(array_as_leaves=False)
-    assert is_treeclass_equal(B, Test(None, None, None, None, static_value("A")))
+    assert is_treeclass_equal(B, Test(None, None, None, None, ("A")))
 
     B = A.at[(A == "a") | (A == "b") | (A == "c")].get(array_as_leaves=False)
-    assert is_treeclass_equal(B, Test(10, 20, 30, None, static_value("A")))
+    assert is_treeclass_equal(B, Test(10, 20, 30, None, ("A")))
 
     B = A.at[(A == "a") | (A == "b") | (A == "c") | (A == "d")].get(
         array_as_leaves=False
     )
-    assert is_treeclass_equal(
-        B, Test(10, 20, 30, jnp.array([1, 2, 3, 4, 5]), static_value("A"))
-    )
+    assert is_treeclass_equal(B, Test(10, 20, 30, jnp.array([1, 2, 3, 4, 5]), ("A")))
 
     @pytc.treeclass
     class L0:
@@ -117,13 +118,10 @@ def test_getter_by_param():
     rhs = L2(10, None, None, L1(1, None, None, L0(1, None, None)))
     assert is_treeclass_equal(lhs, rhs)
 
-    with pytest.raises(NotImplementedError):
+    with pytest.raises(AttributeError):
         t.at["s"].get()
 
-    with pytest.raises(NotImplementedError):
-        t.at["s"].set(100)
-
-    with pytest.raises(NotImplementedError):
+    with pytest.raises(AttributeError):
         t.at["s"].apply(lambda _: 100)
 
 
@@ -134,23 +132,23 @@ def test_getter_by_metadata():
         b: float = field(metadata={"name": "b", "unit": "m"})
         c: float = field(metadata={"name": "c", "unit": "m"})
         d: jnp.ndarray = field(metadata={"name": "d", "unit": "m"})
-        name: str
+        name: str = pytc.static_field()
 
-    A = Test(10, 20, 30, jnp.array([1, 2, 3, 4, 5]), static_value("A"))
+    A = Test(10, 20, 30, jnp.array([1, 2, 3, 4, 5]), ("A"))
 
     B = A.at[A == {"name": "a"}].get(array_as_leaves=False)
-    assert is_treeclass_equal(B, Test(10, None, None, None, static_value("A")))
+    assert is_treeclass_equal(B, Test(10, None, None, None, ("A")))
 
     B = A.at[(A == {"name": "a"}) | (A == {"name": "b"})].get(array_as_leaves=False)
-    assert is_treeclass_equal(B, Test(10, 20, None, None, static_value("A")))
+    assert is_treeclass_equal(B, Test(10, 20, None, None, ("A")))
 
     B = A.at[A == {"": ""}].get(array_as_leaves=False)
-    assert is_treeclass_equal(B, Test(None, None, None, None, static_value("A")))
+    assert is_treeclass_equal(B, Test(None, None, None, None, ("A")))
 
     B = A.at[(A == {"name": "a"}) | (A == {"name": "b"}) | (A == {"name": "c"})].get(
         array_as_leaves=False
     )
-    assert is_treeclass_equal(B, Test(10, 20, 30, None, static_value("A")))
+    assert is_treeclass_equal(B, Test(10, 20, 30, None, ("A")))
 
     B = A.at[
         (A == {"name": "a"})
@@ -158,9 +156,7 @@ def test_getter_by_metadata():
         | (A == {"name": "c"})
         | (A == {"name": "d"})
     ].get(array_as_leaves=False)
-    assert is_treeclass_equal(
-        B, Test(10, 20, 30, jnp.array([1, 2, 3, 4, 5]), static_value("A"))
-    )
+    assert is_treeclass_equal(B, Test(10, 20, 30, jnp.array([1, 2, 3, 4, 5]), ("A")))
 
     @pytc.treeclass
     class L0:
@@ -213,7 +209,7 @@ def test_setter_by_val():
 
     assert is_treeclass_equal(B, C)
 
-    A = Test(10, 20, 30, jnp.array([1, 2, 3, 4, 5]), static_value("A"))
+    A = Test(10, 20, 30, jnp.array([1, 2, 3, 4, 5]), ("A"))
 
     with pytest.raises(AssertionError):
         B = A.at[A].set(0)
@@ -248,22 +244,16 @@ def test_setter_by_val():
 
 
 def test_setter_by_param():
-    A = Test(10, 20, 30, jnp.array([1, 2, 3, 4, 5]), static_value("A"))
+    A = Test(10, 20, 30, jnp.array([1, 2, 3, 4, 5]), ("A"))
 
     B = A.at[A == "a"].set(0)
-    assert is_treeclass_equal(
-        B, Test(0, 20, 30, jnp.array([1, 2, 3, 4, 5]), static_value("A"))
-    )
+    assert is_treeclass_equal(B, Test(0, 20, 30, jnp.array([1, 2, 3, 4, 5]), ("A")))
 
     B = A.at[(A == "a") | (A == "b")].set(0)
-    assert is_treeclass_equal(
-        B, Test(0, 0, 30, jnp.array([1, 2, 3, 4, 5]), static_value("A"))
-    )
+    assert is_treeclass_equal(B, Test(0, 0, 30, jnp.array([1, 2, 3, 4, 5]), ("A")))
 
     B = A.at[(A == "a") | (A == "b") | (A == "c")].set(0)
-    assert is_treeclass_equal(
-        B, Test(0, 0, 0, jnp.array([1, 2, 3, 4, 5]), static_value("A"))
-    )
+    assert is_treeclass_equal(B, Test(0, 0, 0, jnp.array([1, 2, 3, 4, 5]), ("A")))
 
 
 def test_setter_by_metadata():
@@ -275,31 +265,23 @@ def test_setter_by_metadata():
         d: jnp.ndarray = field(metadata={"name": "d", "unit": "m"})
         name: str
 
-    A = Test(10, 20, 30, jnp.array([1, 2, 3, 4, 5]), static_value("A"))
+    A = Test(10, 20, 30, jnp.array([1, 2, 3, 4, 5]), ("A"))
 
     B = A.at[A == {"name": "a"}].set(100, array_as_leaves=False)
-    assert is_treeclass_equal(
-        B, Test(100, 20, 30, jnp.array([1, 2, 3, 4, 5]), static_value("A"))
-    )
+    assert is_treeclass_equal(B, Test(100, 20, 30, jnp.array([1, 2, 3, 4, 5]), ("A")))
 
     B = A.at[(A == {"name": "a"}) | (A == {"name": "b"})].set(
         100, array_as_leaves=False
     )
-    assert is_treeclass_equal(
-        B, Test(100, 100, 30, jnp.array([1, 2, 3, 4, 5]), static_value("A"))
-    )
+    assert is_treeclass_equal(B, Test(100, 100, 30, jnp.array([1, 2, 3, 4, 5]), ("A")))
 
     B = A.at[A == {"": ""}].set(100, array_as_leaves=False)
-    assert is_treeclass_equal(
-        B, Test(10, 20, 30, jnp.array([1, 2, 3, 4, 5]), static_value("A"))
-    )
+    assert is_treeclass_equal(B, Test(10, 20, 30, jnp.array([1, 2, 3, 4, 5]), ("A")))
 
     B = A.at[(A == {"name": "a"}) | (A == {"name": "b"}) | (A == {"name": "c"})].set(
         100, array_as_leaves=False
     )
-    assert is_treeclass_equal(
-        B, Test(100, 100, 100, jnp.array([1, 2, 3, 4, 5]), static_value("A"))
-    )
+    assert is_treeclass_equal(B, Test(100, 100, 100, jnp.array([1, 2, 3, 4, 5]), ("A")))
 
     B = A.at[
         (A == {"name": "a"})
@@ -307,7 +289,7 @@ def test_setter_by_metadata():
         | (A == {"name": "c"})
         | (A == {"name": "d"})
     ].set(100, array_as_leaves=False)
-    assert is_treeclass_equal(B, Test(100, 100, 100, 100, static_value("A")))
+    assert is_treeclass_equal(B, Test(100, 100, 100, 100, ("A")))
 
     @pytc.treeclass
     class L0:
@@ -515,55 +497,31 @@ def test_apply_and_its_derivatives():
     rhs = init.at[init != "a"].divide(2.0)
     assert is_treeclass_equal(lhs, rhs)
 
-    # raise
-    with pytest.raises(ValueError):
-        init.freeze().at[init == "a"].apply(lambda x: x**2)
-
-    with pytest.raises(ValueError):
-        init.freeze().at[(init == "a") | (A == "b")].apply(lambda x: x**2)
-
-    with pytest.raises(ValueError):
-        init.freeze().at[(init == "a") | (A == "b")].set(0)
-
-    with pytest.raises(ValueError):
-        init.freeze().at[(init == init) | (A == "b")].set(0)
-
-    with pytest.raises(ValueError):
-        init.freeze().at[(init == init) | (A == "b")].reduce_sum()
-
     @pytc.treeclass
     class Test:
         a: float = field(metadata={"name": "a", "unit": "m"})
         b: float = field(metadata={"name": "b", "unit": "m"})
         c: float = field(metadata={"name": "c", "unit": "m"})
         d: jnp.ndarray = field(metadata={"name": "d", "unit": "m"})
-        name: str
+        name: str = pytc.static_field()
 
-    A = Test(10, 20, 30, jnp.array([1, 2, 3, 4, 5]), static_value("A"))
+    A = Test(10, 20, 30, jnp.array([1, 2, 3, 4, 5]), ("A"))
 
     B = A.at[A == {"name": "a"}].apply(lambda _: 100, array_as_leaves=False)
-    assert is_treeclass_equal(
-        B, Test(100, 20, 30, jnp.array([1, 2, 3, 4, 5]), static_value("A"))
-    )
+    assert is_treeclass_equal(B, Test(100, 20, 30, jnp.array([1, 2, 3, 4, 5]), ("A")))
 
     B = A.at[(A == {"name": "a"}) | (A == {"name": "b"})].set(
         100, array_as_leaves=False
     )
-    assert is_treeclass_equal(
-        B, Test(100, 100, 30, jnp.array([1, 2, 3, 4, 5]), static_value("A"))
-    )
+    assert is_treeclass_equal(B, Test(100, 100, 30, jnp.array([1, 2, 3, 4, 5]), ("A")))
 
     B = A.at[A == {"": ""}].apply(lambda _: 100, array_as_leaves=False)
-    assert is_treeclass_equal(
-        B, Test(10, 20, 30, jnp.array([1, 2, 3, 4, 5]), static_value("A"))
-    )
+    assert is_treeclass_equal(B, Test(10, 20, 30, jnp.array([1, 2, 3, 4, 5]), ("A")))
 
     B = A.at[(A == {"name": "a"}) | (A == {"name": "b"}) | (A == {"name": "c"})].set(
         100, array_as_leaves=False
     )
-    assert is_treeclass_equal(
-        B, Test(100, 100, 100, jnp.array([1, 2, 3, 4, 5]), static_value("A"))
-    )
+    assert is_treeclass_equal(B, Test(100, 100, 100, jnp.array([1, 2, 3, 4, 5]), ("A")))
 
     B = A.at[
         (A == {"name": "a"})
@@ -571,11 +529,11 @@ def test_apply_and_its_derivatives():
         | (A == {"name": "c"})
         | (A == {"name": "d"})
     ].apply(lambda _: 100, array_as_leaves=False)
-    assert is_treeclass_equal(B, Test(100, 100, 100, 100, static_value("A")))
+    assert is_treeclass_equal(B, Test(100, 100, 100, 100, ("A")))
 
     B = A.at[A != {"name": "a"}].apply(lambda _: 100)
     assert is_treeclass_equal(
-        B, Test(10, 100, 100, jnp.array([100, 100, 100, 100, 100]), static_value("A"))
+        B, Test(10, 100, 100, jnp.array([100, 100, 100, 100, 100]), ("A"))
     )
 
     @pytc.treeclass
@@ -774,3 +732,168 @@ def test_is_leaf():
         t.at[...].apply(lambda x: 10, is_leaf=lambda x: x is None),
         Test([10, 10, 10, 10]),
     )
+
+
+def test_masking():
+    @pytc.treeclass
+    class Linear:
+        weight: jnp.ndarray
+        bias: jnp.ndarray
+
+        def __init__(self, key, in_dim, out_dim):
+            self.weight = jax.random.normal(key, shape=(in_dim, out_dim)) * jnp.sqrt(
+                2 / in_dim
+            )
+            self.bias = jnp.ones((1, out_dim))
+
+        def __call__(self, x):
+            return x @ self.weight + self.bias
+
+    @pytc.treeclass
+    class Dropout:
+        p: float
+        eval: bool | None
+
+        def __init__(self, p: float = 0.5, eval: bool | None = None):
+            """p : probability of an element to be zeroed out"""
+            self.p = p
+            self.eval = eval
+
+        def __call__(self, x, *, key=jr.PRNGKey(0)):
+            return (
+                x
+                if (self.eval is True)
+                else jnp.where(
+                    jr.bernoulli(key, (1 - self.p), x.shape), x / (1 - self.p), 0
+                )
+            )
+
+    @pytc.treeclass
+    class LinearWithDropout:
+        def __init__(self):
+            self.l1 = Linear(key=jr.PRNGKey(0), in_dim=1, out_dim=5)
+            self.d1 = Dropout(p=1.0)  # zero out all elements
+
+        def __call__(self, x):
+            x = self.l1(x)
+            x = self.d1(x)
+            return x
+
+    model = LinearWithDropout()
+    npt.assert_allclose(model(jnp.ones((1, 1))), jnp.array([[0.0, 0.0, 0.0, 0.0, 0.0]]))
+
+    mask = model == "eval"
+    model_no_dropout = model.at[mask].set(True, is_leaf=lambda x: x is None)
+    npt.assert_allclose(
+        model_no_dropout(jnp.ones((1, 1))),
+        jnp.array([[1.2656513, -0.8149204, 0.61661845, 2.7664368, 1.3457328]]),
+    )
+
+
+def test_attribute_get():
+    @pytc.treeclass
+    class l0:
+        a: int = 2
+
+    @pytc.treeclass
+    class Test:
+        a: int = 1
+        b: l0 = l0()
+
+    t = Test()
+    assert t.at["a"].get() == 1
+    assert t.at["b"].at["a"].get() == 2
+
+    with pytest.raises(AttributeError):
+        t.at["a"].at["c"].get()
+
+
+def test_attribute_set():
+    @pytc.treeclass
+    class l0:
+        a: int = 2
+
+    @pytc.treeclass
+    class Test:
+        a: int = 1
+        b: l0 = l0()
+
+    t = Test()
+    t.at["a"].set(10)
+
+    assert is_treeclass_equal(t, Test())
+    assert is_treeclass_equal(t.at["a"].set(10), Test(10, l0()))
+    assert is_treeclass_equal(t.at["b"].at["a"].set(100), Test(1, l0(100)))
+
+    with pytest.raises(AttributeError):
+        t.at["c"].set(10)
+
+    with pytest.raises(AttributeError):
+        t.at["a"].at["c"].set(10)
+
+
+def test_attribute_apply():
+    @pytc.treeclass
+    class l0:
+        a: int = 2
+
+    @pytc.treeclass
+    class Test:
+        a: int = 1
+        b: l0 = l0()
+
+    t = Test()
+    t.at["a"].apply(lambda _: 10)
+
+    assert is_treeclass_equal(t, Test())
+    assert is_treeclass_equal(t.at["a"].apply(lambda _: 10), Test(10))
+    assert is_treeclass_equal(t.at["b"].at["a"].apply(lambda _: 100), Test(1, l0(100)))
+
+    with pytest.raises(AttributeError):
+        t.at["c"].apply(lambda _: 10)
+
+
+def test_method_call():
+    @pytc.treeclass
+    class Test:
+        a: int = 1
+
+        def increment(self):
+            self.a += 1
+
+    t = Test()
+
+    assert is_treeclass_equal(t.at["increment"]()[1], Test(2))
+
+    with pytest.raises(AttributeError):
+        t.at["bla"]()
+
+    with pytest.raises(TypeError):
+        t.at["a"]()
+
+    @pytc.treeclass
+    class A:
+        a: int
+
+        def __call__(self, x):
+            self.a += x
+            return x
+
+    a = A(1)
+    _, b = a.at["__call__"](2)
+
+    assert jtu.tree_leaves(a) == [1]
+    assert jtu.tree_leaves(b) == [3]
+
+
+def test_composed_at():
+    @pytc.treeclass
+    class Test:
+        a: jnp.ndarray = jnp.array([1, 2, 3, 4, 5])
+
+    t = Test()
+
+    assert is_treeclass_equal(t.at[t > 0].at[t < 0].get(), Test(jnp.array([])))
+    assert is_treeclass_equal(
+        t.at[t > 0].at[t == "a"].get(), Test(jnp.array([1, 2, 3, 4, 5]))
+    )
